#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The decoder wrappers for predict synthesis parameters from MIDI or from
MIDI and note expression controls."""

import tensorflow as tf
import ddsp
import ddsp.training
import numpy as np
from .interpretable_conditioning import get_interpretable_conditioning, \
  get_conditioning_dict
from midi_ddsp.data_handling.instrument_name_utils import NUM_INST

tfk = tf.keras
tfkl = tfk.layers


def get_angles(pos, i, d_model):
  angle_rates = 1 / np.power(10000, (2 * (i // 2)) / np.float32(d_model))
  return pos * angle_rates


def positional_encoding(index, d_model):
  angle_rads = get_angles(index,
                          np.arange(d_model)[np.newaxis, :],
                          d_model)

  angle_rads[:, 0::2] = np.sin(angle_rads[:, 0::2])
  angle_rads[:, 1::2] = np.cos(angle_rads[:, 1::2])

  pos_encoding = angle_rads

  return tf.cast(pos_encoding, tf.float32)


class ExpressionMidiDecoder(tfkl.Layer):
  """Decoder wrapper for predicting synthesis parameters from MIDI and
  note expression controls."""

  def __init__(self, decoder, z_preconditioning_stack, multi_instrument=False,
               position_code='none', decoder_type='dilated_conv',
               without_note_expression=False):
    self.decoder_type = decoder_type
    self.z_preconditioning_stack = z_preconditioning_stack
    self.decoder = decoder
    self.position_code = position_code
    self.multi_instrument = multi_instrument
    if multi_instrument:
      self.instrument_emb = tfkl.Embedding(NUM_INST, 64)
    self.without_note_expression = without_note_expression
    super().__init__()

  def gen_params_from_cond(self, conditioning_dict, midi_features,
                           instrument_id=None, synth_params=None,
                           training=False, display_progressbar=False):
    q_pitch, q_vel, f0_loss_weights, onsets, offsets = midi_features
    # note-wise conditioning

    if self.without_note_expression:
      z_conditioning = tf.concat([q_pitch / 127,
                                  tf.cast(onsets, tf.float32)[..., tf.newaxis],
                                  tf.cast(offsets, tf.float32)[..., tf.newaxis]
                                  ], -1)
    else:
      z_conditioning = tf.stop_gradient(
        tf.concat(list(conditioning_dict.values()), -1))
      z_conditioning = tf.concat([z_conditioning,
                                  q_pitch / 127,
                                  tf.cast(onsets, tf.float32)[..., tf.newaxis],
                                  tf.cast(offsets, tf.float32)[..., tf.newaxis]
                                  ], -1)

    if self.position_code == 'index_length':
      note_mask = ddsp.training.nn.get_note_mask_from_onset(q_pitch, onsets)
      each_note_idx = tf.cumsum(note_mask, axis=1) * tf.cast(~(note_mask == 0),
                                                             tf.float32)
      each_note_len = tf.reduce_max(each_note_idx, axis=1,
                                    keepdims=True) * tf.cast(each_note_idx > 0,
                                                             tf.float32)
      each_note_idx = tf.reduce_sum(each_note_idx, axis=-1)[..., tf.newaxis]
      each_note_len = tf.reduce_sum(each_note_len, axis=-1)[..., tf.newaxis]
      relative_position = tf.math.divide_no_nan(each_note_idx, each_note_len)
      z_conditioning = tf.concat([z_conditioning, relative_position], -1)
    elif self.position_code == 'sinusoidal':
      note_mask = ddsp.training.nn.get_note_mask_from_onset(q_pitch, onsets)
      each_note_idx = tf.cumsum(note_mask, axis=1) * tf.cast(~(note_mask == 0),
                                                             tf.float32)
      each_note_idx = tf.reduce_sum(each_note_idx, axis=-1)[..., tf.newaxis]
      pos_code = positional_encoding(each_note_idx.numpy().astype(np.int64), 64)
      z_conditioning = tf.concat([z_conditioning, pos_code], -1)

    # --- Precondition
    z_midi_decoder = self.z_preconditioning_stack(z_conditioning)
    if self.multi_instrument:
      instrument_z = tf.tile(
        self.instrument_emb(instrument_id)[:, tf.newaxis, :],
        [1, z_midi_decoder.shape[1], 1])
      z_midi_decoder = tf.concat([z_midi_decoder, instrument_z], -1)

    # --- MIDI Decoding
    if self.decoder_type == 'dilated_conv':
      params_pred = self.decoder(q_pitch, q_vel, z_midi_decoder)
    elif self.decoder_type == 'noise_dilated_conv':
      noise = tf.random.normal(
        [z_midi_decoder.shape[0], z_midi_decoder.shape[1], 100])
      params_pred = self.decoder(noise, q_pitch, z_midi_decoder)
    elif 'rnn' in self.decoder_type:
      params_pred = self.decoder(q_pitch, z_midi_decoder, conditioning_dict,
                                 out=synth_params, training=training,
                                 display_progressbar=display_progressbar)
    # midi_decoder: [q_pitch, z_midi_decoder] -> synth params
    return z_midi_decoder, params_pred

  def gen_cond_dict(self, synth_params_normalized, midi_features):
    f0, amps, hd, noise = synth_params_normalized
    q_pitch, q_vel, f0_loss_weights, onsets, offsets = midi_features

    f0_midi_gt = ddsp.core.midi_to_hz(q_pitch, midi_zero_silence=True)
    conditioning = get_interpretable_conditioning(f0_midi_gt, f0, amps, hd,
                                                  noise)

    # --- Z Note Encoding
    conditioning_dict = get_conditioning_dict(conditioning, q_pitch, onsets,
                                              pool_type='note_pooling')
    return conditioning_dict

  def call(self, features, synth_params_normalized, midi_features,
           training=False, synth_params=None):
    conditioning_dict = self.gen_cond_dict(synth_params_normalized,
                                           midi_features)

    instrument_id = features['instrument_id'] if self.multi_instrument else None
    if self.decoder_type == 'rnn_f0_ld':
      synth_params = features
    z_midi_decoder, params_pred = self.gen_params_from_cond(conditioning_dict,
                                                            midi_features,
                                                            instrument_id=
                                                            instrument_id,
                                                            synth_params=
                                                            synth_params,
                                                            training=training)

    params_pred['z_midi_decoder'] = z_midi_decoder

    return conditioning_dict, params_pred


class MidiDecoder(tfkl.Layer):
  """Decoder wrapper for predicting synthesis parameters from only MIDI."""

  def __init__(self, decoder, multi_instrument=False):
    super().__init__()
    self.decoder = decoder
    self.multi_instrument = multi_instrument
    if multi_instrument:
      self.instrument_emb = tfkl.Embedding(NUM_INST, 64)
    self.pitch_emb = tfkl.Embedding(128, 64)

  def call(self, features, synth_params_normalized, midi_features,
           training=False, synth_params=None):
    instrument_id = features['instrument_id'] if self.multi_instrument else None
    q_pitch, q_vel, f0_loss_weights, onsets, offsets = midi_features
    instrument_z = tf.tile(self.instrument_emb(instrument_id)[:, tf.newaxis, :],
                           [1, q_pitch.shape[1], 1])
    z_midi_decoder = tf.concat(
      [self.pitch_emb(tf.cast(q_pitch, tf.int64)[..., 0]),
       tf.cast(onsets, tf.float32)[..., tf.newaxis],
       tf.cast(offsets, tf.float32)[..., tf.newaxis],
       instrument_z],
      -1)  # HACK
    params_pred = self.decoder(q_pitch, z_midi_decoder, out=features,
                               training=training)
    return {}, params_pred



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Loss helper for handling and logging GAN loss."""

import tensorflow as tf
from midi_ddsp.modules.loss_helper import LossHelper
import ddsp
from ddsp.spectral_ops import F0_RANGE


class GANLossHelper(LossHelper):
  """Loss helper for handling and logging GAN loss.

  Args:
    lambda_recon: the reconstruction coefficient.
    lambda_feat: the feature matching loss coefficient.
    lambda_G: the generator loss coefficient.
    sg_z: whether to stop gradient on the feature sequence input to the
      Synthesis Generator.
  """

  def __init__(self, lambda_recon=1, lambda_feat=10, lambda_G=1, sg_z=True):
    super().__init__()
    self.loss_list = [
      'disc_loss',
      'gen_loss',
      'fm_loss'
    ]
    self.n_layers_D = 4
    self.num_D = 3
    self.lambda_feat = lambda_feat
    self.lambda_recon = lambda_recon
    self.lambda_G = lambda_G
    self.sg_z = sg_z
    self.init_metrics()

  def compute_disc_loss(self, D_fake_det, D_real):
    loss_D = 0
    for scale in D_fake_det:
      loss_D += tf.reduce_mean(scale[-1] ** 2)

    for scale in D_real:
      loss_D += tf.reduce_mean((scale[-1] - 1) ** 2)

    return {'disc_loss': loss_D}

  def compute_gen_loss(self, D_fake, D_real, total_loss_recon):
    loss_G = 0
    for scale in D_fake:
      loss_G += tf.reduce_mean((scale[-1] - 1) ** 2)

    loss_feat = 0
    feat_weights = 4.0 / (self.n_layers_D + 1)
    D_weights = 1.0 / self.num_D
    wt = D_weights * feat_weights
    for i in range(self.num_D):
      for j in range(len(D_fake[i]) - 1):
        loss_feat += wt * tf.reduce_mean(
          tf.abs(D_fake[i][j] - tf.stop_gradient(D_real[i][j])))  # MAE

    fm_loss = self.lambda_feat * loss_feat

    total_loss = self.lambda_recon * total_loss_recon + \
                 self.lambda_G * loss_G + fm_loss

    return total_loss, {'gen_loss': loss_G, 'fm_loss': fm_loss}

  def get_disc_input(self, outputs):
    if self.sg_z:
      cond = tf.stop_gradient(outputs['params_pred']['z_midi_decoder'])
    else:
      cond = outputs['params_pred']['z_midi_decoder']

    real_outputs = tf.concat([
      ddsp.core.hz_to_midi(outputs['synth_params']['f0_hz'] / F0_RANGE),
      outputs['synth_params']['amplitudes'],
      outputs['synth_params']['harmonic_distribution'],
      outputs['synth_params']['noise_magnitudes']
    ], -1)
    real_outputs = tf.stop_gradient(real_outputs)

    fake_outputs = tf.concat([
      ddsp.core.hz_to_midi(outputs['params_pred']['f0_hz'] / F0_RANGE),
      outputs['params_pred']['amplitudes'],
      outputs['params_pred']['harmonic_distribution'],
      outputs['params_pred']['noise_magnitudes']
    ], -1)

    return cond, real_outputs, fake_outputs



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Expression generator module class."""

from abc import ABC
import tensorflow as tf
from ddsp.training import nn
from midi_ddsp.data_handling.instrument_name_utils import NUM_INST
from midi_ddsp.modules.cond_rnn import TwoLayerCondAutoregRNN

tfk = tf.keras
tfkl = tfk.layers


class LangModelOutputLayer(tfkl.Layer):
  def __init__(self, n_out, nhid=256):
    super().__init__()
    self.n_out = n_out
    self.dense_out = nn.FcStackOut(ch=nhid, layers=2, n_out=self.n_out)

  def call(self, inputs):
    outputs = self.dense_out(inputs)
    outputs = {'raw_output': outputs}
    return outputs


class ExpressionGenerator(TwoLayerCondAutoregRNN, tf.keras.Model, ABC):
  """Expression Generator that takes note sequence as input and predicts
  note expression controls."""

  # TODO：(yusongwu) merge teacher_force and autoregressive function,
  # things need to change:
  # def sample_out(self, out, cond, time, training=False):
  # curr_out = self.sample_out(curr_out, cond, i, training=training)
  # output = self.split_teacher_force_output(output, cond)
  def __init__(self, n_out=6, nhid=128, norm=True, dropout=0.5):
    super().__init__(
      nhid=nhid,
      n_out=n_out,
      input_dropout=True,
      input_dropout_p=0.5,
      dropout=dropout,
    )
    self.birnn = tfkl.Bidirectional(tfkl.GRU(
      units=nhid, return_sequences=True, dropout=dropout
    ), )
    self.dense_out = LangModelOutputLayer(nhid=nhid, n_out=self.n_out)
    self.norm = nn.Normalize('layer') if norm else None
    self.pitch_emb = tfkl.Embedding(128, 64)
    self.duration_emb = tfkl.Dense(64)
    self.instrument_emb = tfkl.Embedding(NUM_INST, 64)

  def autoregressive(self, cond, training=False):
    note_pitch = cond['note_pitch'][..., tf.newaxis]
    cond = self.encode_cond(cond, training=training)
    batch_size = cond.shape[0]
    length = cond.shape[1]
    prev_out = tf.tile([[0.0]], [batch_size, self.n_out])[:, tf.newaxis,
               :]  # go_frame
    prev_states = (None, None)
    overall_outputs = []

    for i in range(length):
      curr_cond = cond[:, i, :][:, tf.newaxis, :]
      prev_out = self.encode_out(prev_out)
      curr_out, curr_states = self._one_step(curr_cond, prev_out, prev_states,
                                             training=training)
      curr_out = self.sample_out(curr_out,
                                 note_pitch[:, i, :][:, tf.newaxis, :])
      overall_outputs.append(curr_out)
      prev_out, prev_states = curr_out['output'], curr_states

    outputs = {}
    for k in curr_out.keys():
      outputs[k] = tf.concat([x[k] for x in overall_outputs], 1)
    return outputs

  def teacher_force(self, cond, out, training=True):
    note_pitch = cond['note_pitch'][..., tf.newaxis]
    out_shifted = self.right_shift_encode_out(out)
    cond = self.encode_cond(cond, training=training)
    z_in = tf.concat([cond, out_shifted], -1)
    z_out, *states = self.rnn1(z_in, training=training)
    z_out, *states = self.rnn2(z_out, training=training)
    output = self.decode_out(z_out)
    output = self.sample_out(output, note_pitch)
    return output

  def encode_cond(self, cond, training=False):
    z_pitch = self.pitch_emb(cond['note_pitch'])
    z_duration = self.duration_emb(cond['note_length'])
    z_instrument = self.instrument_emb(
      tf.tile(cond['instrument_id'][:, tf.newaxis], [1, z_pitch.shape[1]]))
    cond = tf.concat([z_pitch, z_duration, z_instrument], -1)
    cond = self.birnn(cond, training=training)
    return cond

  def decode_out(self, z_out):
    if self.norm is not None:
      z_out = self.norm(z_out)
    output = self.dense_out(z_out)
    return output

  def sample_out(self, out, note_pitch):
    output = out.copy()
    sampled_output = out['raw_output']
    rest_note_mask = tf.cast(note_pitch != 0, tf.float32)
    sampled_output *= rest_note_mask
    output['output'] = sampled_output
    return output

  def call(self, cond, out=None, training=False):

    if training:
      outputs = self.teacher_force(cond, out, training=training)
    else:
      outputs = self.autoregressive(cond, training=training)
    return outputs


def get_fake_data_expression_generator(target_dim):
  instrument_id = tf.ones([1], dtype=tf.int64)
  cond = {
    'note_pitch': tf.ones([1, 32], dtype=tf.int64),
    'note_length': tf.ones([1, 32, 1], dtype=tf.float32),
    'instrument_id': instrument_id
  }
  target = tf.ones([1, 32, target_dim], dtype=tf.float32)
  fake_data = {
    'cond': cond,
    'target': target
  }
  return fake_data



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The reverb modules used in MIDI-DDSP."""

import tensorflow as tf
import ddsp

tfk = tf.keras
tfkl = tfk.layers


def get_exp_decay(total_length, start_length, decay_exponent):
  """Get an exponential decay from 1 to 0.
  Args:
    total_length: Total length of the decay.
    start_length: The point where the decay starts.
    decay_exponent: The coefficient of the exponential decay.

  Returns: decay, the exponential decay from 1 to 0.

  """
  time = tf.linspace(0.0, 1.0, total_length - start_length)
  decay = tf.exp(-decay_exponent * time)
  decay = tf.concat([tf.ones(start_length), decay], 0)
  return decay


class ReverbModules(tfkl.Layer):
  """A reverb module for multi-instrument training.
  An unique reverb parameter is used for each instrument id.
  """

  def __init__(self, num_reverb=1, reverb_length=48000):
    super().__init__()
    self.num_reverb = num_reverb
    self.reverb_length = reverb_length
    self.reverb = ddsp.effects.Reverb(trainable=False,
                                      reverb_length=reverb_length)
    initializer = tf.random_normal_initializer(mean=0, stddev=1e-6)
    self.magnitudes_embedding = tfkl.Embedding(num_reverb, reverb_length,
                                               embeddings_initializer=
                                               initializer)

  def call(self, audio, reverb_number=0, training=False):
    batch_size = audio.shape[0]
    if isinstance(reverb_number, int):
      reverb_number = tf.repeat(tf.constant([reverb_number], dtype=tf.int64),
                                batch_size)
    if self.num_reverb == 1 or reverb_number is None:  # unified reverb
      reverb_number = tf.repeat(tf.constant([0], dtype=tf.int64), batch_size)
    ir_magnitudes = self.magnitudes_embedding(reverb_number)
    if not training:
      ir_magnitudes = ir_magnitudes * get_exp_decay(
        self.reverb_length, 16000, 4)[tf.newaxis, ...]
    output = self.reverb(audio, ir_magnitudes)
    return output



"""Utility functions for inference."""
#  Copyright 2021 The DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Lint as: python3

import ddsp
import numpy as np
import pandas as pd
import tensorflow as tf

# Note expression keys. This list is strictly ordered.
CONDITIONING_KEYS = ['amplitude_mean',
                     'amplitude_std',
                     'vibrato_extend',
                     'brightness',
                     'attack_level',
                     'amplitudes_max_pos']


def to_length(data, dst_length=1000, axis=1):
  """Make data to dst_length.
  Will pad 0 if have less length, and will cut off if have more length."""
  data_length = data.shape[axis]
  before_shape = data.shape[:axis]
  after_shape = data.shape[axis + 1:]
  if data_length < dst_length:
    pad_data = tf.zeros([*before_shape, dst_length - data_length, *after_shape],
                        dtype=data.dtype)
    return tf.concat([data, pad_data], axis)
  else:
    size = list(data.shape)
    size[axis] = dst_length
    return tf.slice(data, [0] * len(data.shape), size)


def ensure_same_length(data_all, axis=1):
  """Ensure the same length for all the data in the list."""
  max_length = max([d.shape[axis] for d in data_all])
  return [to_length(d, dst_length=max_length, axis=axis) for d in data_all]


def make_same_length(src, dst):
  """Make src the same length as the dst."""
  length = dst.shape[1]
  if src.shape[1] > length:
    return src[:, :length, ...]
  else:
    pad_shape = [src.shape[0], length - src.shape[1]] + src.shape[2:]
    return tf.concat([src, tf.zeros(pad_shape, dtype=src.dtype)], 1)


def conditioning_df_to_midi_features(conditioning_df, length=None):
  """Convert conditioning_df to midi features."""
  total_length = length if length is not None else \
    conditioning_df.tail(1)['offset'].values[0]
  q_pitch = np.zeros((1, total_length, 1), dtype=np.float32)
  onsets = np.zeros((1, total_length), dtype=np.int64)
  offsets = np.zeros((1, total_length), dtype=np.int64)
  for _, note in conditioning_df.iterrows():
    on = int(note['onset'])
    off = int(note['offset'])
    q_pitch[:, on:off + 1, :] = note['pitch']
    onsets[:, on] = 1
    offsets[:, off - 1] = 1

  # q_pitch, q_vel, f0_loss_weight, onsets, offsets
  midi_features = (q_pitch, q_pitch, q_pitch, onsets, offsets)
  return midi_features


def expression_generator_output_to_conditioning_df(expression_generator_outputs,
                                                   expression_generator_cond,
                                                   frame_size=0.004):
  """
  Convert expression generator outputs to conditioning_df.
  Args:
      expression_generator_outputs: The output of the expression generator.
      expression_generator_cond: (in dataset data['cond'])
      frame_size: the frame size, in second.

  Returns:

  """
  expression_generator_outputs = scale_expression_generator_output(
    expression_generator_outputs[0])
  conditioning_dict_keys = CONDITIONING_KEYS
  data_all = []
  onset = 0
  for i in range(expression_generator_outputs.shape[0]):
    pitch = expression_generator_cond['note_pitch'][0, i].numpy()
    length = int(round(expression_generator_cond['note_length'][
                         0, i, 0].numpy() / frame_size))
    offset = onset + length
    data_all.append(
      expression_generator_outputs[i].numpy().tolist() + [pitch, onset, offset,
                                                          length])
    onset += length
  columns = conditioning_dict_keys + ['pitch', 'onset', 'offset', 'note_length']
  conditioning_df = pd.DataFrame(data_all, columns=columns)
  return conditioning_df


def conditioning_df_to_expression_generator_output(conditioning_df,
                                                   instrument_id,
                                                   frame_size=0.004):
  """Convert conditioning_df to expression generator output."""
  note_pitch = conditioning_df['pitch'].values
  note_length = conditioning_df['note_length'].values

  note_pitch = tf.convert_to_tensor(note_pitch)[tf.newaxis, ...]
  note_length = tf.convert_to_tensor(note_length)[tf.newaxis, ...]
  note_length = tf.cast(note_length, tf.float32)[..., tf.newaxis] * frame_size

  note_sequence = {'note_pitch': note_pitch, 'note_length': note_length,
                   'instrument_id': tf.constant([instrument_id])}
  return note_sequence


def scale_expression_generator_output(output):
  """Scale expression generator output."""
  scale = np.ones((1, output.shape[-1]))
  scale = tf.convert_to_tensor(scale, tf.float32)
  output /= scale
  return output


def conditioning_df_to_audio(synthesis_generator, conditioning_df,
                             instrument_id, length=None,
                             display_progressbar=False):
  """Synthesize audio from conditioning_df."""
  conditioning_dict = conditioning_df_to_dict(conditioning_df, length=length)

  midi_features = conditioning_df_to_midi_features(conditioning_df,
                                                   length=length)
  midi_audio, midi_control_params, midi_synth_params = synthesis_generator. \
    gen_audio_from_cond_dict(conditioning_dict,
                             midi_features,
                             instrument_id=instrument_id,
                             display_progressbar=display_progressbar)

  return midi_audio, midi_control_params, midi_synth_params


def conditioning_dict_to_df(conditioning_dict, onsets, offsets, q_pitch):
  """Convert conditioning_dict to conditioning_df."""
  # Add onset to the beginning and offset to the end if needed.
  onsets, offsets, q_pitch = onsets.numpy(), offsets.numpy(), q_pitch.numpy()
  if q_pitch.ndim == 3:
    q_pitch = q_pitch[..., 0]
  onsets = np.where(onsets[0] == 1)[0]
  offsets = np.where(offsets[0] == 1)[0]
  if (len(onsets) == 0) or (
        q_pitch[0, 0] != 0 and onsets[0] != 0 and offsets[0] != 0):
    onsets = np.append(0, onsets)
  if (len(offsets) == 0) or (
        q_pitch[0, -1] != 0 and offsets[-1] != q_pitch.shape[1] and offsets[
    -1] != q_pitch.shape[1] - 1):
    offsets = np.append(offsets, q_pitch.shape[1])
  if offsets[0] == 0:
    offsets = offsets[1:]
  assert len(onsets) == len(offsets)

  # Make conditioning_df.
  data_all = []
  for on, off in zip(onsets.tolist(), offsets.tolist()):
    data_all.append(
      [v[0, on, 0].numpy() for v in conditioning_dict.values()] + [
        q_pitch[0, on], on, off,
        off - on])  # max(off-on, 1)
  columns = list(conditioning_dict.keys()) + ['pitch', 'onset', 'offset',
                                              'note_length']
  conditioning_df = pd.DataFrame(data_all, columns=columns)
  return conditioning_df


def conditioning_df_to_dict(conditioning_df, length=None):
  """Convert conditioning_df to conditioning_dict."""
  total_length = length if length is not None else \
    conditioning_df.tail(1)['offset'].values[0]
  conditioning_names = list(conditioning_df.columns)[:-4]
  conditioning_dict = {k: np.zeros((1, total_length, 1), dtype=np.float32) for k
                       in conditioning_names}
  for _, note in conditioning_df.iterrows():
    for name in conditioning_names:
      on = int(note['onset'])
      off = int(note['offset'])
      conditioning_dict[name][:, on:off + 1, :] = note[name]
  conditioning_dict = {k: tf.convert_to_tensor(v) for k, v in
                       conditioning_dict.items()}
  return conditioning_dict


def get_process_group(n_frames, frame_size=64, sample_rate=16000,
                      use_angular_cumsum=True):
  harmonic_synth = ddsp.synths.Harmonic(n_frames * frame_size, sample_rate,
                                        use_angular_cumsum=use_angular_cumsum)
  noise_synth = ddsp.synths.FilteredNoise(n_frames * frame_size, sample_rate)
  add = ddsp.processors.Add(name='add')
  # Create ProcessorGroup.
  dag = [(harmonic_synth, ['amplitudes', 'harmonic_distribution', 'f0_hz']),
         (noise_synth, ['noise_magnitudes']),
         (add, ['filtered_noise/signal', 'harmonic/signal'])]

  processor_group = ddsp.processors.ProcessorGroup(dag=dag,
                                                   name='processor_group')
  return processor_group



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""The dataloader for expression generator."""

import tensorflow as tf
import numpy as np
from itertools import cycle

_AUTOTUNE = tf.data.experimental.AUTOTUNE


def random_choose_data(data):
  def gen():
    for piece in cycle(data):
      j = np.random.randint(0, len(piece))
      yield piece[j]

  return gen


def get_lang_model_dataset(data_language_model_preprocessed, batch_size=256,
                           repeats=-1):
  # np.random.shuffle(data_language_model_preprocessed)
  output_types = {
    'note_pitch': tf.int64,
    'note_length': tf.int64,
    'conditioning_feature': tf.float32,
    'instrument_id': tf.int64,
  }
  padded_shapes = {
    'note_pitch': [None],
    'note_length': [None],
    'conditioning_feature': [None, None],
    'instrument_id': [],
  }

  def _reshape_tensors(data):
    cond = {
      'note_pitch': data['note_pitch'],
      'note_length': tf.cast(data['note_length'], tf.float32)[
                       ..., tf.newaxis] * 0.004,
      'instrument_id': data['instrument_id']
    }
    target = data['conditioning_feature']
    scale = np.ones((1, 6))  # HACK
    scale = tf.convert_to_tensor(scale, tf.float32)
    target *= scale

    return {
      'cond': cond,
      'target': target,
    }

  if repeats != 1:
    data_iter = random_choose_data(data_language_model_preprocessed)
    dataset = tf.data.Dataset.from_generator(data_iter,
                                             output_types=output_types)

  else:
    dataset = tf.data.Dataset.from_generator(
      lambda: data_language_model_preprocessed, output_types=output_types)

  dataset = dataset.padded_batch(batch_size, padded_shapes=padded_shapes)
  dataset = dataset.repeat(repeats)
  dataset = dataset.map(_reshape_tensors, num_parallel_calls=_AUTOTUNE)
  dataset = dataset.prefetch(tf.data.experimental.AUTOTUNE)

  return dataset



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Install midi-ddsp."""

import setuptools

with open('README.md') as f:
  long_description = f.read()

setuptools.setup(
  name='midi-ddsp',
  version='0.2.2',
  description='Synthesis of MIDI with DDSP',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author='Google Inc. & Yusong Wu',
  author_email='no-reply@google.com, wuyusongwys@gmail.com',
  url='http://github.com/magenta/midi-ddsp',
  license='Apache 2.0',
  packages=setuptools.find_packages(),
  scripts=[],
  install_requires=[
    'ddsp==3.2.0',
    'pretty_midi',
    'music21',
    'pandas',
    # (yusongwu) Force overwriting tfds-nightly as tfds-nightly
    # will raise error under windows.
    'tensorflow_datasets'
  ],
  extras_require={
    'test': ['pytest', 'pylint!=2.5.0'],
  },
  entry_points={
    'console_scripts': [
      'midi_ddsp_synthesize = midi_ddsp.midi_ddsp_synthesize:main',
      'midi_ddsp_download_model_weights = midi_ddsp.download_model_weights:main'
    ],
  },
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: Apache Software License',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
  ],
  tests_require=['pytest'],
  setup_requires=['pytest-runner'],
  keywords='audio MIDI MIDI-synthesizer machinelearning music',
)



"""Utility functions used in training loop."""

#  Copyright 2021 The DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#  Lint as: python3

import numpy as np
import inspect
import tensorflow as tf
import random
import os
import argparse
import librosa
from ast import literal_eval
import matplotlib.pyplot as plt

from midi_ddsp.utils.audio_io import save_wav


def plot_spec(wav, sr, title='', vmin=-8, vmax=1, save_path=None):
  """Plot and save the spectrogram."""
  spec = np.log(np.abs(librosa.stft(wav, n_fft=512 + 256)))
  librosa.display.specshow(spec, sr=sr, vmin=vmin, vmax=vmax, cmap='magma')
  plt.title(title)
  if save_path:
    plt.savefig(save_path)
    plt.close()


def get_hp(file):
  """Retrieve hyperparameters from log files."""
  with open(file, 'r') as f:
    log_all = f.readlines()
    f.close()
  hp = None
  for line in log_all:  # (yusongwu) hack here
    if '{\'add_synth_loss\':' in line:
      hp = literal_eval(line.strip()[32:])
      break
  if hp is None:
    raise RuntimeError('Failed to load the model.')
  return hp


def set_seed(seed):
  """Set the random seed."""
  os.environ['PYTHONHASHSEED'] = str(seed)
  random.seed(seed)
  np.random.seed(seed)
  tf.random.set_seed(seed)
  os.environ['TF_DETERMINISTIC_OPS'] = '1'


def print_hparams(hp):
  """Print the hyperparameters"""
  attributes = inspect.getmembers(hp, lambda a: not inspect.isroutine(a))
  return {a[0]: a[1] for a in attributes if
          not (a[0].startswith('__') and a[0].endswith('__'))}


def save_results(pred_batch, ref_wav_batch, log_dir, save_folder_name,
                 sample_rate, acoustic_params=None,
                 separate_signal=None, file_name=None, bottleneck_feature=None,
                 note=None, no_pred_suffix=False):
  """Save the evaluation results."""
  if file_name and pred_batch.shape[0] > 1:
    raise RuntimeError('Got a file name to save, but has batch_size > 1')
  pred = pred_batch.numpy()
  if ref_wav_batch is not None:
    ref_wav = ref_wav_batch.numpy()
  if log_dir is None:
    save_dir = save_folder_name
  else:
    save_dir = os.path.join(log_dir,
                            f'results_{os.path.basename(save_folder_name)}')
  os.makedirs(save_dir, exist_ok=True)
  output_features_save_dir = os.path.join(save_dir, 'output_features')
  separate_signal_save_dir = os.path.join(save_dir, 'separate_signal')
  spectrogram_dir = os.path.join(save_dir, 'spectrogram')
  os.makedirs(output_features_save_dir, exist_ok=True)
  os.makedirs(separate_signal_save_dir, exist_ok=True)
  os.makedirs(spectrogram_dir, exist_ok=True)
  for i in range(pred.shape[0]):

    save_file_name = file_name if file_name else i

    pred_file_suffix = '' if no_pred_suffix else '_pred'
    save_wav(pred[i],
             os.path.join(save_dir, f'{save_file_name}{pred_file_suffix}.wav'),
             sample_rate=sample_rate)
    plot_spec(pred[i], sr=sample_rate,
              title=f'{save_file_name}{pred_file_suffix}.wav',
              save_path=os.path.join(spectrogram_dir,
                                     f'{save_file_name}{pred_file_suffix}.png'))
    if ref_wav_batch is not None:
      save_wav(ref_wav[i], os.path.join(save_dir, f'{save_file_name}_ref.wav'),
               sample_rate=sample_rate)
      plot_spec(ref_wav[i], sr=sample_rate, title=f'{save_file_name}_ref.wav',
                save_path=os.path.join(spectrogram_dir,
                                       f'{save_file_name}_ref.png'))
    if acoustic_params:
      for k in acoustic_params.keys():
        np.save(
          os.path.join(output_features_save_dir, f'{save_file_name}_{k}.npy'),
          acoustic_params[k][i].numpy())
    if separate_signal:
      for k in separate_signal.keys():
        save_wav(separate_signal[k][i].numpy(),
                 os.path.join(separate_signal_save_dir,
                              f'{save_file_name}_{k}.wav'),
                 sample_rate=sample_rate)
    if bottleneck_feature:
      for k in bottleneck_feature.keys():
        if 'loss' not in k and 'perplexity' not in k:
          np.save(
            os.path.join(output_features_save_dir, f'{save_file_name}_{k}.npy'),
            bottleneck_feature[k][i].numpy())
    if note:
      for k in note.keys():
        np.save(
          os.path.join(output_features_save_dir, f'{save_file_name}_{k}.npy'),
          note[k][i].numpy())


def str2bool(v):
  """Enable boolean in argparse by passing string."""
  # https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse
  if isinstance(v, bool):
    return v
  if v.lower() in ('yes', 'true', 't', 'y', '1'):
    return True
  elif v.lower() in ('no', 'false', 'f', 'n', '0'):
    return False
  else:
    raise argparse.ArgumentTypeError('Boolean value expected.')



#  Copyright 2022 The MIDI-DDSP Authors.
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      http://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Loss helper for handling and logging the losses."""

import tensorflow as tf


class LossHelper:
  """Loss helper for handling and logging the losses."""
  def __init__(self):
    self.metrics = None

  def init_metrics(self):
    self.metrics = {f'mean_{l}': tf.keras.metrics.Mean(name=f'mean_{l}') for l
                    in self.loss_list}

  def reset_metrics(self):
    for m in self.metrics.values():
      m.reset_states()

  def update_metrics(self, loss_dict):
    for loss_name, value in loss_dict.items():
      self.metrics[f'mean_{loss_name}'](value)

  def compute_loss(self, inputs, outputs, **kwargs):
    return self.loss_group(inputs, outputs, **kwargs)

  def get_loss_log(self):
    log = ''
    for loss_name in self.loss_list:
      loss_result = self.metrics[f'mean_{loss_name}'].result()
      log = log + f'| {loss_name} {loss_result:5.4f} '
    return log

  def write_mean_summary(self, writer, group_name, step):
    with writer.as_default():
      for loss_name in self.loss_list:
        loss_result = self.metrics[f'mean_{loss_name}'].result().numpy()
        tf.summary.scalar(f'{group_name}/{loss_name}', loss_result, step)

  @staticmethod
  def write_summary(loss_dict, writer, group_name, step):
    with writer.as_default():
      for loss_name, value in loss_dict.items():
        tf.summary.scalar(f'{group_name}/{loss_name}', value.numpy(), step)
